/* Copyright (C) 2015-2020 |Meso|Star> (contact@meso-star.com)
 *
 * This software is a computer program whose purpose is to describe a
 * virtual 3D environment that can be ray-traced and sampled both robustly
 * and efficiently.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef S3D_MESH_H
#define S3D_MESH_H

#include "s3d_c.h"

#include "s3d_buffer.h"
#include "s3d_geometry.h"

#include <rsys/dynamic_array_u32.h>
#include <rsys/dynamic_array_float.h>
#include <rsys/ref_count.h>

/* Generate the index buffer data type */
#define BUFFER_NAME index_buffer
#define BUFFER_DARRAY darray_u32
#include "s3d_buffer.h"

/* Generate the vertex buffer data type */
#define BUFFER_NAME vertex_buffer
#define BUFFER_DARRAY darray_float
#include "s3d_buffer.h"

struct mesh { /* Triangular mesh */
  struct index_buffer* indices;
  struct vertex_buffer* attribs[S3D_ATTRIBS_COUNT__];
  enum s3d_type attribs_type[S3D_ATTRIBS_COUNT__];
  struct darray_float cdf;
  struct hit_filter filter;

  struct s3d_device* dev;
  ref_T ref;
};

extern LOCAL_SYM res_T
mesh_create
  (struct s3d_device* dev,
   struct mesh** mesh);

extern LOCAL_SYM void
mesh_ref_get
  (struct mesh* mesh);

extern LOCAL_SYM void
mesh_ref_put
  (struct mesh* mesh);

extern LOCAL_SYM void
mesh_clear
  (struct mesh* mesh);

extern LOCAL_SYM size_t
mesh_get_ntris
  (const struct mesh* mesh);

extern LOCAL_SYM size_t
mesh_get_nverts
  (const struct mesh* mesh);

extern LOCAL_SYM uint32_t*
mesh_get_ids
  (struct mesh* mesh);

extern LOCAL_SYM float*
mesh_get_pos
  (struct mesh* mesh);

extern LOCAL_SYM float*
mesh_get_attr
  (struct mesh* mesh,
   const enum s3d_attrib_usage usage);

extern LOCAL_SYM float
mesh_compute_area
  (struct mesh* mesh);

/* Compute mesh CDF */
extern LOCAL_SYM res_T
mesh_compute_cdf
  (struct mesh* mesh);

extern LOCAL_SYM float
mesh_compute_volume
  (struct mesh* mesh,
   const char flip_surface); /* Flip the shape surface or not */

extern LOCAL_SYM res_T
mesh_setup_indexed_vertices
  (struct mesh* mesh,
   const unsigned ntris,
   void (*get_indices)(const unsigned itri, unsigned ids[3], void* ctx),
   const unsigned nverts,
   struct s3d_vertex_data attribs[],
   const unsigned nattribs,
   void* data);

extern LOCAL_SYM void
mesh_compute_aabb
  (struct mesh* mesh,
   float lower[3],
   float upper[3]);

extern LOCAL_SYM void
mesh_copy_indexed_vertices
  (const struct mesh* src,
   struct mesh* dst);

#endif /* S3D_MESH_H */

